/*

  Copyright (c) 2009, VTT Technical Research Center of Finland
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  * Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
  * Neither the name of the VTT Technical Research Center of Finland nor the
  names of its contributors may be used to endorse or promote products
  derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY COPYRIGHT HOLDERS ''AS IS'' AND ANY
  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/**
 * \file consumer_kp.c
 *
 * \brief A template implementation for Knowledge Processor that obtains sensor
 *  values from SIB and prints the values to the terminal.
 *
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <pthread.h> /* thread for handling the subscribe indications */
#include <unistd.h>

#include <kpilow/kpi_low.h>
#include <kpilow/sskp_errno.h>

#define MAX_SENSORS 3

typedef struct sensor
{
  char uri[SS_SUBJECT_MAX_LEN];
  char name[SS_OBJECT_MAX_LEN];
  char location[SS_OBJECT_MAX_LEN];
  char unit[SS_OBJECT_MAX_LEN];
  char value[SS_OBJECT_MAX_LEN];

  struct sensor * next;
}sensor_t;

typedef struct thread_param
{
  ss_info_t ss_info;
  ss_subs_info_t * subs_info;
}thread_param_t;

ss_info_t ss_info; /* pointer to the smart space info */

void print_sensors(sensor_t * sensors, int number);
int query_sensors(ss_info_t * ss_info, sensor_t * sensor, int number);
int query_sensor_info(ss_info_t * ss_info, sensor_t * sensors);
int subscribe_sensor_values(ss_info_t * ss_info, ss_subs_info_t * subs_info, sensor_t * sensors, int number);
void * subscribe_handler(void * parameters);

void print_sensors(sensor_t * sensors, int number)
{
  int i = 0;

  for(i = 0; i < number; i++)
    {
      printf("Sensor [URI = %s]\n", sensors[i].uri);
      printf("  Name: %s\n", sensors[i].name);
      printf("  Location: %s\n", sensors[i].location);
      printf("  Unit type:  %s\n", sensors[i].unit);
      printf("  Value: %s\n", sensors[i].value);
    }
}

int query_sensors(ss_info_t * ss_info, sensor_t * sensors, int number)
{
  ss_triple_t * sensor_rqst = NULL;
  ss_triple_t * first_triple = NULL;
  ss_triple_t * triple = NULL;
  int i;

  ss_add_triple(&sensor_rqst, ss_info->space_id, "hasSensor", SS_RDF_SIB_ANY, SS_RDF_TYPE_URI, SS_RDF_TYPE_URI);
  if(ss_query(ss_info, sensor_rqst, &first_triple) < 0)
    {
      printf("Unable to query sensors\n");
      return -1;
    }

  ss_delete_triples(sensor_rqst);

  triple = first_triple;
  for(i = 0; i < number && triple; i++)
    {
      strcpy(sensors[i].uri, triple->object);
      triple = triple->next;
    }

  ss_delete_triples(first_triple);

  return i;
}

int query_sensor_info(ss_info_t * ss_info, sensor_t * sensor)
{
  ss_triple_t * sensor_info_rqst = NULL;
  ss_triple_t * result_triple = NULL;

  /* multiple triple queries do not currently work in the CSIB */

  ss_add_triple(&sensor_info_rqst, sensor->uri, "hasName", SS_RDF_SIB_ANY, SS_RDF_TYPE_URI, SS_RDF_TYPE_URI);
  if(ss_query(ss_info, sensor_info_rqst, &result_triple) < 0)
  {
    printf("Unable to query name info\n");
    //return -1;
  }
  else strcpy(sensor->name, result_triple->object);
  ss_delete_triples(sensor_info_rqst);
  sensor_info_rqst = NULL;
  ss_delete_triples(result_triple);
  result_triple = NULL;

  ss_add_triple(&sensor_info_rqst, sensor->uri, "hasLocation", SS_RDF_SIB_ANY, SS_RDF_TYPE_URI, SS_RDF_TYPE_URI);
  if(ss_query(ss_info, sensor_info_rqst, &result_triple) < 0)
  {
    printf("Unable to query location info\n");
    //return -1;
  }
  else strcpy(sensor->location, result_triple->object);
  ss_delete_triples(sensor_info_rqst);
  sensor_info_rqst = NULL;

  ss_delete_triples(result_triple);
  result_triple = NULL;

  ss_add_triple(&sensor_info_rqst, sensor->uri, "hasUnitType", SS_RDF_SIB_ANY, SS_RDF_TYPE_URI, SS_RDF_TYPE_URI);
  if(ss_query(ss_info, sensor_info_rqst, &result_triple) < 0)
  {
    printf("Unable to query unit type info\n");
    //return -1;
  }
  else strcpy(sensor->unit, result_triple->object);

  ss_delete_triples(sensor_info_rqst);
  sensor_info_rqst = NULL;
  ss_delete_triples(result_triple);
  result_triple = NULL;

  return 0;
}

int subscribe_sensor_value(ss_info_t * ss_info, ss_subs_info_t * subs_info, sensor_t * sensor)
{
  pthread_t thread;
  thread_param_t * param;
  ss_triple_t * triple_rqst = NULL;
  ss_triple_t * triple = NULL;

  ss_add_triple(&triple_rqst, sensor->uri, "hasValue", SS_RDF_SIB_ANY, SS_RDF_TYPE_URI, SS_RDF_TYPE_URI);

  /* subscribe to values of all sensors found from Smart Space */
  if(ss_subscribe(ss_info, subs_info, triple_rqst, &triple) < 0)
  {
    printf("Failed to subscribe\n");
    return -1;
  }

  ss_delete_triples(triple_rqst);

  strcpy(sensor->value, triple->object);   
  ss_delete_triples(triple);

  param = (thread_param_t *)malloc(sizeof(thread_param_t));
  param->ss_info = *ss_info;
  param->subs_info = subs_info;

  if(pthread_create(&thread, NULL, subscribe_handler, (void *)param))
  {
    printf("ERROR: unable to create subscribe handling thread.\n");
    return 0;
  }

  return 1;
}

void * subscribe_handler(void * parameters)
{
  thread_param_t * param = (thread_param_t *)parameters;
  ss_info_t ss_info = param->ss_info; /* When the subscribe indications are evaluated parallel in a thread it is neccessary to make copy of the ss_info struct to avoid situations where data is overwritten */
  ss_subs_info_t * subs_info = param->subs_info; 
  ss_triple_t * n_val = NULL;
  ss_triple_t * o_val = NULL;
  ss_triple_t * triple = NULL;
  int status;
 
  while(1)
  {
    status = ss_subscribe_indication(&ss_info, subs_info, &n_val, &o_val, 1000);

    if(status == 0)
      continue; /* timeout */
    if(status < 0)
    {
      ss_perror(ss_info.ss_errno);
      continue;
    }
    if(status == 1)
    {
      printf("Received subscribe indication [ID = %s]\n", subs_info->id);

      triple = n_val;
      while(triple)
      {
        printf("Sensor:%s\n", triple->subject);
        printf(" New value:  %s\n", triple->object);
        triple = triple->next;
      }
      triple = o_val;
      while(triple)
      {
        printf("Sensor:%s\n", triple->subject);
        printf(" Obsolete value:  %s\n", triple->object);
        triple = triple->next;
      }

      ss_delete_triples(n_val);
      ss_delete_triples(o_val);
      n_val = NULL;
      o_val = NULL;
    }
    else
    {
      printf("Unsubscribe indication received\n");
      break;
    }
  }

  pthread_exit(NULL);
}

int main()
{
  char input[10] = {0};
  int sensor_number;
  ss_subs_info_t subs_info[MAX_SENSORS];
  int i = 0;

  ss_discovery(&ss_info);

  sensor_t sensors[MAX_SENSORS];

  printf("Space ID = %s\n", ss_info.space_id);

  /* Join to the Smart Space */
  if(ss_join(&ss_info, "Consumer KP") != -1)
  {
    printf("Joined successfully\n");

    sensor_number = query_sensors(&ss_info, sensors, MAX_SENSORS);
    if(sensor_number > 0)
    {
      for(i=0; i < sensor_number; i++)
      {
        query_sensor_info(&ss_info, &sensors[i]);
        subscribe_sensor_value(&ss_info, &subs_info[i], &sensors[i]);
      }
      print_sensors(sensors, sensor_number);

      while(1)
      {
        printf("\nWrite exit to quit\n");
        scanf("%s", input);
        if(!strcmp("exit", input))
          break;
      }

      for(i=0; i < sensor_number; i++)
      {
        printf("Unsubscribing [ID=%s]\n", subs_info[i].id);
        ss_unsubscribe(&ss_info, &subs_info[i]);
        usleep(1000000); // this is needed because SIB cannot process many simultaneus messages
      }
    }
    else
      printf("No sensors found from the Smart Space\n");

    ss_leave(&ss_info);

  }
  else
  {
    printf("ERROR: unable to join the Smart Space\n");
    ss_perror(ss_info.ss_errno);
  }
  return 0;
}
